import requests
import json


class JSONRPCError(Exception):
    def __init__(self, code, message):
        super().__init__()
        self._code = code
        self._message = message

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    def __str__(self):
        return f"JSONRPCError (code: {self.code}, message: {self.message})"


class BSHLocalAPI:
    def __init__(self, controller_ip: str, certificate, key):
        self._certificate = certificate
        self._key = key
        self._controller_ip = controller_ip
        self._api_root = f"https://{self._controller_ip}:8444/smarthome"
        self._rpc_root = f"https://{self._controller_ip}:8444/remote/json-rpc"

        # Settings for all API calls
        self._requests_session = requests.Session()
        self._requests_session.cert = (self._certificate, self._key)
        self._requests_session.headers.update({"api-version": "1.0", "Content-Type": "application/json"})
        self._requests_session.verify = False

        import urllib3
        urllib3.disable_warnings()

    def _get_api_result_or_fail(self, api_url, expected_type=None, expected_element_type=None, headers=None):
        result = self._requests_session.get(api_url, headers=headers)
        if not result.ok:
            self._process_nok_result(result)

        else:
            if len(result.content) > 0:
                result = json.loads(result.content)
                if expected_type is not None:
                    assert result['@type'] == expected_type
                if expected_element_type is not None:
                    for result_ in result:
                        assert result_['@type'] == expected_element_type

                return result
            else:
                return {}

    def _put_api_or_fail(self, api_url, body):
        result = self._requests_session.put(api_url, data=json.dumps(body))
        if not result.ok:
            self._process_nok_result(result)
        if len(result.content) > 0:
            return json.loads(result.content)
        else:
            return {}

    def _post_api_or_fail(self, api_url, body):
        result = self._requests_session.post(api_url, data=json.dumps(body))
        if not result.ok:
            self._process_nok_result(result)
        if len(result.content) > 0:
            return json.loads(result.content)
        else:
            return {}

    def _process_nok_result(self, result):
        print(f"Body: {result.request.body}")
        print(f"Headers: {result.request.headers}")
        print(f"URL: {result.request.url}")
        raise ValueError(f"API call returned non-OK result (code {result.status_code})!: {result.content}")

    # API calls here
    def get_rooms(self):
        api_url = f"{self._api_root}/rooms"
        return self._get_api_result_or_fail(api_url, expected_element_type="room")

    def get_devices(self):
        api_url = f"{self._api_root}/devices"
        return self._get_api_result_or_fail(api_url, expected_element_type="device")

    def get_device_service(self, device_id, service_id):
        api_url = f"{self._api_root}/devices/{device_id}/services/{service_id}"
        return self._get_api_result_or_fail(api_url, expected_type="DeviceServiceData")

    def put_device_service_state(self, device_id, service_id, state_update):
        api_url = f"{self._api_root}/devices/{device_id}/services/{service_id}/state"
        self._put_api_or_fail(api_url, state_update)

    def long_polling_subscribe(self):
        data = [
            {
                "jsonrpc": "2.0",
                "method": "RE/subscribe",
                "params": ["com/bosch/sh/remote/*", None]
            }
        ]
        result = self._post_api_or_fail(self._rpc_root, data)
        assert result[0]["jsonrpc"] == "2.0"
        if "error" in result[0].keys():
            raise JSONRPCError(result[0]['error']['code'], result[0]['error']['message'])
        else:
            return result[0]["result"]

    def long_polling_poll(self, poll_id, wait_seconds=30):
        data = [
            {
                "jsonrpc": "2.0",
                "method": "RE/longPoll",
                "params": [poll_id, wait_seconds]
            }
        ]
        result = self._post_api_or_fail(self._rpc_root, data)
        assert result[0]["jsonrpc"] == "2.0"
        if "error" in result[0].keys():
            raise JSONRPCError(result[0]['error']['code'], result[0]['error']['message'])
        else:
            return result[0]["result"]

    def long_polling_unsubscribe(self, poll_id):
        data = [
            {
                "jsonrpc": "2.0",
                "method": "RE/unsubscribe",
                "params": [poll_id]
            }
        ]
        result = self._post_api_or_fail(self._rpc_root, data)
        assert result[0]["jsonrpc"] == "2.0"
        if "error" in result[0].keys():
            raise JSONRPCError(result[0]['error']['code'], result[0]['error']['message'])
        else:
            return result[0]["result"]


import typing
import logging
import threading
import time
import sys

from .api import BSHLocalAPI, JSONRPCError
from .device import BSHLocalDevice
from .room import BSHLocalRoom
from .services_impl import SUPPORTED_DEVICE_SERVICE_IDS

logger = logging.getLogger("bshlocal")


class BSHLocalSession:
    def __init__(self, controller_ip: str, certificate, key):
        # API
        self._api = BSHLocalAPI(controller_ip=controller_ip, certificate=certificate, key=key)

        # Subscription status
        self._poll_id = None

        # All devices
        self._rooms_by_id = {}
        self._devices_by_id = {}

        self._enumerate_devices()
        self._enumerate_rooms()

        self._polling_thread = None
        self._stop_polling_thread = False

    def _enumerate_devices(self):
        raw_devices = self._api.get_devices()
        for raw_device in raw_devices:
            device_id = raw_device['id']

            if set(raw_device['deviceServiceIds']).isdisjoint(SUPPORTED_DEVICE_SERVICE_IDS):
                logger.info(f"Skipping device id {device_id} which has no services that are supported by this library")
                continue

            device = BSHLocalDevice(api=self._api, raw_device=raw_device)
            self._devices_by_id[device_id] = device

    def _enumerate_rooms(self):
        raw_rooms = self._api.get_rooms()
        for raw_room in raw_rooms:
            room_id = raw_room["id"]
            room = BSHLocalRoom(api=self._api, raw_room=raw_room)
            self._rooms_by_id[room_id] = room

    def _long_poll(self, wait_seconds=30):
        if self._poll_id is None:
            self._poll_id = self.api.long_polling_subscribe()
            logger.debug(f"Subscribed for long poll. Poll id: {self._poll_id}")
        try:
            raw_results = self.api.long_polling_poll(self._poll_id, wait_seconds)
            for raw_result in raw_results:
                self._process_long_polling_poll_result(raw_result)

            return True
        except JSONRPCError as json_rpc_error:
            if json_rpc_error.code == -32001:
                self._poll_id = None
                logger.warning(f"SHC claims unknown poll id. Invalidating poll id and trying resubscribe next time...")
                return False
            else:
                raise json_rpc_error

    def _maybe_unsubscribe(self):
        if self._poll_id is not None:
            self.api.long_polling_unsubscribe(self._poll_id)

    def _process_long_polling_poll_result(self, raw_result):
        assert raw_result["@type"] == "DeviceServiceData"
        device_id = raw_result["deviceId"]
        if device_id in self._devices_by_id.keys():
            device = self._devices_by_id[device_id]
            device.process_long_polling_poll_result(raw_result)
        else:
            logger.debug(f"Skipping polling result with unknown device id {device_id}.")

    def start_polling(self):
        if self._polling_thread is None:
            def polling_thread_main():
                while not self._stop_polling_thread:
                    try:
                        if not self._long_poll():
                            logging.warning("_long_poll returned False. Waiting 1 second.")
                            time.sleep(1.0)
                    except RuntimeError as err:
                        if self._stop_polling_thread:
                            logging.info("Stopping polling thread after expected runtime error.")
                            return
                        else:
                            logging.error(f"Runtime error in running polling thread: {err}. Waiting 15 seconds.")
                            time.sleep(15.0)
                    except Exception as ex:
                        logging.error(f"Error in polling thread: {ex}. Waiting 15 seconds.")
                        time.sleep(15.0)

            self._polling_thread = threading.Thread(target=polling_thread_main, name="BSHLocalPollingThread")
            self._polling_thread.start()

        else:
            raise ValueError("Already polling!")

    def stop_polling(self):
        if self._polling_thread is not None:
            self._stop_polling_thread = True
            self._polling_thread.join()

            self._maybe_unsubscribe()
            self._polling_thread = None
        else:
            raise ValueError("Not polling!")

    @property
    def devices(self) -> typing.Sequence[BSHLocalDevice]:
        return list(self._devices_by_id.values())

    def device(self, device_id) -> BSHLocalDevice:
        return self._devices_by_id[device_id]

    @property
    def rooms(self) -> typing.Sequence[BSHLocalRoom]:
        return list(self._rooms_by_id.values())

    def room(self, room_id) -> BSHLocalRoom:
        return self._rooms_by_id[room_id]

    @property
    def api(self):
        return self._api



from enum import Enum
from .device_service import BSHLocalDeviceService


class TemperatureLevelService(BSHLocalDeviceService):
    @property
    def temperature(self) -> float:
        return float(self.state["temperature"])

    def summary(self):
        super().summary()
        print(f"    Temperature              : {self.temperature}")


class RoomClimateControlService(BSHLocalDeviceService):
    class OperationMode(Enum):
        AUTOMATIC = "AUTOMATIC"
        MANUAL = "MANUAL"

    @property
    def operation_mode(self) -> OperationMode:
        return self.OperationMode(self.state["operationMode"])

    @operation_mode.setter
    def operation_mode(self, value: OperationMode):
        self.put_state_element("operationMode", value.value)

    @property
    def setpoint_temperature(self) -> float:
        return float(self.state["setpointTemperature"])

    @setpoint_temperature.setter
    def setpoint_temperature(self, value: float):
        self.put_state_element("setpointTemperature", value)

    @property
    def setpoint_temperature_eco(self) -> float:
        return float(self.state["setpointTemperatureForLevelEco"])

    @property
    def setpoint_temperature_comfort(self) -> float:
        return float(self.state["setpointTemperatureForLevelComfort"])

    @property
    def ventilation_mode(self) -> bool:
        return self.state["ventilationMode"]

    @property
    def low(self) -> bool:
        return self.state["low"]
    
    @low.setter
    def low(self, value: bool):
        self.put_state_element("low", value)

    @property
    def boost_mode(self) -> bool:
        return self.state["boostMode"]
    
    @boost_mode.setter
    def boost_mode(self, value: bool):
        self.put_state_element("boostMode", value)

    @property
    def summer_mode(self) -> bool:
        return self.state["summerMode"]

    @property
    def supports_boost_mode(self) -> bool:
        return self.state["supportsBoostMode"]

    @property
    def show_setpoint_temperature(self) -> bool:
        if "showSetpointTemperature" in self.state.keys():
            return self.state["showSetpointTemperature"]
        else:
            return False

    def summary(self):
        super().summary()
        print(f"    Operation Mode           : {self.operation_mode}")
        print(f"    Setpoint Temperature     : {self.setpoint_temperature}")
        print(f"    Setpoint Temperature ECO : {self.setpoint_temperature_eco}")
        print(f"    Setpoint Temperature CMF : {self.setpoint_temperature_comfort}")
        print(f"    Ventilation Mode         : {self.ventilation_mode}")
        print(f"    Low                      : {self.low}")
        print(f"    Boost Mode               : {self.boost_mode}")
        print(f"    Summer Mode              : {self.summer_mode}")
        print(f"    Supports Boost Mode      : {self.supports_boost_mode}")
        print(f"    Show Setpoint Temperature: {self.show_setpoint_temperature}")


class ShutterContactService(BSHLocalDeviceService):
    class State(Enum):
        CLOSED = "CLOSED"
        OPEN = "OPEN"

    @property
    def value(self) -> State:
        return self.State(self.state["value"])

    def summary(self):
        super().summary()
        print(f"    Value                    : {self.value}")


class ValveTappetService(BSHLocalDeviceService):
    @property
    def position(self) -> int:
        return int(self.state["position"])

    def summary(self):
        super().summary()
        print(f"    Position                 : {self.position}")


SERVICE_MAPPING = {"TemperatureLevel": TemperatureLevelService,
                   "RoomClimateControl": RoomClimateControlService,
                   "ShutterContact": ShutterContactService,
                   "ValveTappet": ValveTappetService}

SUPPORTED_DEVICE_SERVICE_IDS = SERVICE_MAPPING.keys()


def build(api, raw_device_service):
    device_service_id = raw_device_service["id"]
    assert device_service_id in SUPPORTED_DEVICE_SERVICE_IDS, "Device service is supported"
    return SERVICE_MAPPING[device_service_id](api=api, raw_device_service=raw_device_service)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bshlocal",
    version="0.0.5",
    author="Clemens-Alexander Brust",
    author_email="cabrust@pm.me",
    description="Bosch Smart Home Local API Python Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cabrust/bshlocal",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['requests==2.22']
)